#include "para.h"
static struct stat_client *clients;


/* This must exactly match para.h */
const char *status_item_list[NUM_STAT_ITEMS] = {
	"status_bar",		"status",		"num_played",
	"mtime",		"length_min",		"length_sec",
	"file_size",		"status_flags",		"format",
	"score",		"audio_file_info1",	"audio_file_info2",
	"audio_file_info3",	"dbinfo1",		"dbinfo2",
	"dbinfo3",		"decoder_flags",	"audiod_status",
	"play_time",		"uptime",		"offset",
	"length",		"stream_start",		"current_time",
	"audiod_uptime",	"dbtool"
};
#define FOR_EACH_STAT_CLIENT(p) for ((p) = clients; (p); (p) = (p)->next)
#define FOR_EACH_STAT_ITEM(i) for (i = 0; i < NUM_STAT_ITEMS; i++)

static void dump_stat_client_list(void)
{
	struct stat_client *p;
	int i = 0;

	PARA_DEBUG_LOG("%s", "dumping list\n");
	FOR_EACH_STAT_CLIENT(p)
		PARA_INFO_LOG("client #%d: fd %d\n", i++, p->fd);
	PARA_DEBUG_LOG("%s", "list dumped\n");
}

int stat_client_add(int fd)
{
	struct stat_client *p, *new_client = para_malloc(sizeof(struct stat_client));
	int num_clients = 1;

	PARA_INFO_LOG("adding client on fd %d\n", fd);
	new_client->next = NULL;
	new_client->fd = fd;
	if (!clients) {
		/* very first client */
		clients = new_client;
		goto out;
	}
	FOR_EACH_STAT_CLIENT(p) {
		num_clients++;
		if (p->next)
			continue;
		p->next = new_client;
		goto out;
	}
out:
	PARA_INFO_LOG("client #%d on fd %d added\n", num_clients, fd);
	dump_stat_client_list();
	return num_clients;
}

int dump_stat_line(char *line)
{
	struct stat_client *p, *prev = NULL; 
	char *buf = make_message("%s\n", line);
	ssize_t len = strlen(buf);
	int num_clients = 0;

	FOR_EACH_STAT_CLIENT(p) {
		int fd, ret;
//		PARA_DEBUG_LOG("%s: p=%lx\n", __func__, (long)p);
		fd = p->fd;
		ret = write(fd, buf, len);
		PARA_DEBUG_LOG("dumped %s to fd %d, ret = %d\n", line, fd, ret);
		if (ret == len ) {
			num_clients++;
			prev = p;
			continue;
		}
		/* write error */
		close(fd);
		PARA_INFO_LOG("deleting client on fd %d\n", fd);
		if (!prev) {
			PARA_INFO_LOG("%s", "last client disconnected\n");
			/* last client disconnected */
			clients = NULL;
			goto out;
		}
		prev->next = p->next;
		PARA_DEBUG_LOG("deleted client on fd %d, p=%lx\n", fd,
			(long) p);
		free(p);
		p = prev;
		dump_stat_client_list();
	}
out:
	free(buf);
	PARA_DEBUG_LOG("%d client(s)\n", num_clients);
	return num_clients;
}

int dump_empty_status(void)
{
	int i;
	char *empty_items = para_strdup(NULL);

	FOR_EACH_STAT_ITEM(i) {
		char *tmp = make_message("%s:\n", status_item_list[i]);
		empty_items = para_strcat(empty_items, tmp);
		free(tmp);
	}
//	PARA_DEBUG_LOG("%s: empty items: %s\n", __func__, empty_items);
	i = dump_stat_line(empty_items);
	free(empty_items);
	return i;
}

/*
 * check if line starts with known command.
 */
int stat_line_valid(const char *line)
{
	int i;
	size_t line_len;

	if (!line || !*line)
		return -1;
	line_len = strlen(line);
	for (i = 0; i < NUM_STAT_ITEMS; i++) {
		const char *s = status_item_list[i];
		size_t item_len = strlen(s);

		if (line_len > item_len && line[item_len] == ':' &&
				!strncmp(line, s, item_len))
			return i;
	}
	return -1;
}

/* if line handler is NULL, return number of complete lines in buf.
 * Otherwise, call line_handler for each complete line, but no more than
 * num times. Copy rest of buffer (last chunk containing incomplete line)
 * to start of buffer and return its size.
 */
int check_buf_for_items(char *buf, int n, void (*line_handler)(char *), int num)
{
	char *start = buf, *end;
	int i, num_lines = 0;

	while (start < buf + n) {
		char *next_null;
		char *next_cr;

		next_cr = memchr(start, '\n', buf + n - start);
		next_null = memchr(start, '\0', buf + n - start);
		if (!next_cr && !next_null)
			break;
		if (next_cr && next_null) {
			end = next_cr < next_null? next_cr : next_null;
		} else if (next_null) {
			end = next_null;
		} else
			end = next_cr;
		num_lines++;
		if (line_handler) {
			*end = '\0';
			line_handler(start);
			start = ++end;
			if (num && num_lines >= num)
				break;
		} else
			start = ++end;
	}
	if (!line_handler)
		return num_lines;
	i = buf + n - start;
	if (i && i != n)
		memmove(buf, start, i);
	return i;
}

